/* Copyright (c) 1999-2001 by Juliusz Chroboczek.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE. */

#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/SimpleP.h>
#include <X11/Xmu/Converters.h>
#include <X11/Xatom.h>
#include <X11/Xmu/Atoms.h>
#include "UGridP.h"

#define DefaultCellWidth(fgw) (((fgw)->ugrid.font->max_bounds.width) + \
                               ((fgw)->ugrid.internal_pad * 2))
#define DefaultCellHeight(fgw) ((fgw)->ugrid.font->ascent + \
                                (fgw)->ugrid.font->descent + \
                                ((fgw)->ugrid.internal_pad * 2))

#define CellWidth(fgw) (((int)(fgw)->core.width-(fgw)->ugrid.grid_width) / \
                        (fgw)->ugrid.cell_cols - \
                        (fgw)->ugrid.grid_width)
#define CellHeight(fgw) (((int)(fgw)->core.height-(fgw)->ugrid.grid_width) / \
                         (fgw)->ugrid.cell_rows - \
                         (fgw)->ugrid.grid_width)

#define MIN(a, b) ((a)<=(b)?(a):(b))
#define MAX(a, b) ((a)<=(b)?(b):(a))


#define offset(field) XtOffsetOf(UGridRec, field)

static XtResource resources[] = {
    {XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
     offset(ugrid.foreground_pixel), XtRString, XtDefaultForeground},
    {XtNgridColor, XtCForeground, XtRPixel, sizeof(Pixel),
     offset(ugrid.grid_pixel), XtRString, XtDefaultForeground},
    {XtNfont,  XtCFont, XtRFontStruct, sizeof(XFontStruct *),
     offset(ugrid.font),XtRString, XtDefaultFont},
    {XtNcellColumns, XtCCellColumns, XtRInt, sizeof(int),
     offset(ugrid.cell_cols), XtRImmediate, (XtPointer)0},
    {XtNcellRows, XtCCellRows, XtRInt, sizeof(int),
     offset(ugrid.cell_rows), XtRImmediate, (XtPointer)0},
    {XtNcellWidth, XtCCellWidth, XtRInt, sizeof(int),
     offset(ugrid.cell_width), XtRImmediate, (XtPointer)0},
    {XtNcellHeight, XtCCellHeight, XtRInt, sizeof(int),
     offset(ugrid.cell_height), XtRImmediate, (XtPointer)0},
    {XtNstartCodepoint, XtCStartCodepoint, XtRDimension, sizeof(Dimension),
     offset(ugrid.start_codepoint), XtRImmediate, (XtPointer)0},
    {XtNinternalPad, XtCInternalPad, XtRInt, sizeof(int),
     offset(ugrid.internal_pad), XtRImmediate, (XtPointer)4},
    {XtNgridWidth, XtCGridWidth, XtRInt, sizeof(int),
     offset(ugrid.grid_width), XtRImmediate, (XtPointer)1},
    {XtNcallback, XtCCallback, XtRCallback, sizeof(XtPointer),
     offset(ugrid.callbacks), XtRCallback, (XtPointer)NULL},
};

#undef offset
  
static void Initialize(Widget, Widget, ArgList, Cardinal*);
static void Realize(Widget, Mask*, XSetWindowAttributes *attributes);
static void Resize(Widget w);
static void Redisplay(Widget, XEvent*, Region);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal*);
static void ClassInitialize(void);
static void Destroy(Widget);
static XtGeometryResult QueryGeometry(Widget, 
                                      XtWidgetGeometry*, XtWidgetGeometry*);
static void PageForward(Widget, XEvent*, String*, Cardinal*);
static void PageBackward(Widget, XEvent*, String*, Cardinal*);
static void Notify(Widget, XEvent*, String*, Cardinal*);
static void SelectCodepoint(Widget, XEvent*, String*, Cardinal*);
static void ClearSelection(Widget, XEvent*, String*, Cardinal*);
static Boolean ConvertSelection(Widget, Atom*, Atom *, Atom *, 
                                XtPointer*, unsigned long*, int *);
static void LoseSelection(Widget, Atom*);
static void InsertSelection(Widget, XEvent*, String*, Cardinal*);
static void SelectionReceived(Widget, XtPointer, Atom*, Atom*,
                              XtPointer, unsigned long *, int *);

static char defaultTranslations[] = 
"\
<Btn1Up>:       SelectCodepoint()\n\
<Btn2Up>:       InsertSelection()\n\
<Key>n:         PageForward()\n\
<Key>Right:     PageForward()\n\
<Key>p:         PageBackward()\n\
<Key>Left:      PageBackward()\n\
<Key>Delete:    ClearSelection()\n\
<Key>BackSpace: ClearSelection()\
";

static XtActionsRec actions_list[] = {
    {"SelectCodepoint", SelectCodepoint},
    {"PageForward", PageForward},
    {"PageBackward", PageBackward},
    {"ClearSelection", ClearSelection},
    {"InsertSelection", InsertSelection},
};


UGridClassRec ugridClassRec = {
    {
      /* core_class fields */   
      /* superclass             */      (WidgetClass) &simpleClassRec,
      /* class_name             */      "UGrid",
      /* widget_size            */      sizeof(UGridRec),
      /* class_initialize       */      ClassInitialize,
      /* class_part_initialize  */      NULL,
      /* class_inited           */      FALSE,
      /* initialize             */      Initialize,
      /* initialize_hook        */      NULL,
      /* realize                */      Realize,
      /* actions                */      actions_list,
      /* num_actions            */      XtNumber(actions_list),
      /* resources              */      resources,
      /* num_resources          */      XtNumber(resources),
      /* xrm_class              */      NULLQUARK,
      /* compress_motion        */      TRUE,
      /* compress_exposure      */      TRUE,
      /* compress_enterleave    */      TRUE,
      /* visible_interest       */      FALSE,
      /* destroy                */      Destroy,
      /* resize                 */      Resize,
      /* expose                 */      Redisplay,
      /* set_values             */      SetValues,
      /* set_values_hook        */      NULL,
      /* set_values_almost      */      XtInheritSetValuesAlmost,
      /* get_values_hook        */      NULL,
      /* accept_focus           */      NULL,
      /* version                */      XtVersion,
      /* callback_private       */      NULL,
      /* tm_table               */      defaultTranslations,
      /* query_geometry         */      QueryGeometry,
      /* display_accelerator    */      XtInheritDisplayAccelerator,
      /* extension              */      NULL
    },
/* Simple class fields initialization */
    {
        /* change_sensitive     */      XtInheritChangeSensitive
    },
/* UGrid class fields initialization */
    {
        /* ignore               */      0
    }
};
WidgetClass ugridWidgetClass = (WidgetClass)&ugridClassRec;


static GC get_gc (UGridWidget ugw, Pixel fore, Pixel back)
{
    XtGCMask mask;
    XGCValues gcv;

    mask = (GCForeground | GCBackground | GCFunction | GCFont);
    gcv.foreground = fore;
    gcv.background = back;
    gcv.function = GXcopy;
    gcv.font = ugw->ugrid.font->fid;
    if (ugw->ugrid.grid_width > 0) {
        mask |= GCLineWidth;
        gcv.line_width = ((ugw->ugrid.grid_width < 2) ? 0 : 
                          ugw->ugrid.grid_width);
    }
    return (XtGetGC ((Widget) ugw, mask, &gcv));
}

static void 
ClassInitialize()
{
    XtAddConverter(XtRString, XtRLong, XmuCvtStringToLong, NULL, 0);
}

static void 
Initialize(Widget request, Widget new, ArgList args, Cardinal *num_args)
{
    UGridWidget requg = (UGridWidget)request;
    UGridWidget newug = (UGridWidget)new;
    XFontStruct *fs = newug->ugrid.font;

    if(requg->ugrid.cell_cols<=0)
        newug->ugrid.cell_cols = 16;
    if(requg->ugrid.cell_rows<=0)
        newug->ugrid.cell_rows = 16;

    if(requg->ugrid.cell_width<=0)
        newug->ugrid.cell_width = fs ? DefaultCellWidth(newug) : 1;
    if(requg->ugrid.cell_height<=0)
        newug->ugrid.cell_height = fs ? DefaultCellHeight(newug) : 1;
  
    if(newug->core.width<=0)
        newug->core.width=
            newug->ugrid.cell_width * newug->ugrid.cell_cols +
            newug->ugrid.grid_width*(newug->ugrid.cell_cols+1);

    if(newug->core.height<=0)
        newug->core.height=
            newug->ugrid.cell_height * newug->ugrid.cell_rows +
            newug->ugrid.grid_width*(newug->ugrid.cell_rows+1);

    newug->ugrid.selected_codepoint = NO_CODEPOINT;
}

static void
Realize(Widget w, Mask *valueMask, XSetWindowAttributes *attributes)
{
    UGridWidget ugw = (UGridWidget)w;
    UGridPart *p = &ugw->ugrid;

    p->text_gc = get_gc(ugw, p->foreground_pixel, ugw->core.background_pixel);
    p->selected_gc = get_gc(ugw, ugw->core.background_pixel, p->foreground_pixel);
    p->grid_gc = get_gc(ugw, p->grid_pixel, ugw->core.background_pixel);
    Resize(w);

    (XtSuperclass(w)->core_class.realize)(w, valueMask, attributes);
    return;
}

static void
redisplay_rectangle(UGridWidget ugw, XRectangle *rect)
{
    UGridPart *p = &ugw->ugrid;
    Display *dpy = XtDisplay(ugw);
    Window win = XtWindow(ugw);
    unsigned long codepoint;
    XChar2b codechar;
  
    int i, j, mini, minj, maxi, maxj;
    int ch, cw, hgw, hgw1, xoff, yoff;
  
    cw = p->cell_width + p->grid_width;
    ch = p->cell_height + p->grid_width;
    hgw = p->grid_width / 2;      /* half grid width, biased downwards */
    hgw1 = (p->grid_width+1) / 2; /* biased upwards */

    xoff = (cw-p->font->max_bounds.width)/2;
    yoff = (ch-(p->font->ascent+p->font->descent))/2;
    
    /* determine which cells to repaint */
    if(rect == NULL) {
        mini = 0;
        minj = 0;
        maxi = p->cell_cols;
        maxj = p->cell_rows;
    } else {
        mini = MAX(0, rect->x/cw);
        minj = MAX(0, rect->y/ch);
        maxi = MIN(p->cell_cols,
                   (rect->x + p->grid_width + rect->width + cw-1) / cw);
        maxj = MIN(p->cell_rows,
                   (rect->y + p->grid_width + rect->height + ch-1) / ch);
    }

    /* repaint the grid */
    if(p->grid_width > 0) {
        for(i=mini; i<=maxi; i++)
            XDrawLine(dpy, win, p->grid_gc,
                      hgw+i*cw, 0, hgw+i*cw, 2*hgw+p->cell_rows*ch);
        for(j=minj; j<=maxj; j++)
            XDrawLine(dpy, win, p->grid_gc,
                      0, hgw+j*ch, 2*hgw+p->cell_cols*cw, hgw+j*ch);
    }

    /* repaint the cells */
    for(j=0; j<p->cell_rows; j++) {
        for(i=0; i<p->cell_cols; i++) {
            codepoint = 
                (long)p->start_codepoint + i + j * (long)p->cell_cols;
            if(codepoint < 0 || codepoint > 0xFFFF)
                continue;
            codechar.byte1 = ((codepoint>>8) & 0xFF);
            codechar.byte2 = (codepoint & 0xFF);
            if(codepoint != p->selected_codepoint) {
                XDrawString16(dpy, win, p->text_gc, 
                              hgw1 + i*cw + xoff, 
                              hgw1 + j*ch + p->font->ascent + yoff, 
                              &codechar, 1);
            } else {
                XFillRectangle(dpy, win, p->text_gc,
                               p->grid_width + i*cw, p->grid_width + j*ch, 
                               cw - p->grid_width, ch - p->grid_width);
                XDrawString16(dpy, win, p->selected_gc,
                              hgw1 + i*cw+xoff, 
                              hgw1 + j*ch + p->font->ascent + yoff, 
                              &codechar, 1);
            }
        }
    }
}

static void
Redisplay(Widget w, XEvent *event, Region region)
{
    UGridWidget ugw = (UGridWidget)w;
    XRectangle rect;
    
    XClipBox(region, &rect);
    redisplay_rectangle(ugw, &rect);
}

static void
Resize(Widget w)
{
    UGridWidget ugw=(UGridWidget)w;

    ugw->ugrid.cell_width = CellWidth(ugw);
    if(ugw->ugrid.cell_width <= 0)
        ugw->ugrid.cell_width = 1;
    ugw->ugrid.cell_height = CellHeight(ugw);
    if(ugw->ugrid.cell_height <= 0)
        ugw->ugrid.cell_height = 1;
}

static Boolean
SetValues(Widget current, Widget request, Widget new, 
          ArgList args, Cardinal *num_args)
{
    UGridWidget curug = (UGridWidget)current;
    UGridWidget newug = (UGridWidget)new;
    Bool redisplay = FALSE;
    
    if(curug->ugrid.font != newug->ugrid.font ||
       curug->ugrid.internal_pad != newug->ugrid.internal_pad) {
        newug->ugrid.cell_width = DefaultCellWidth(newug);
        newug->ugrid.cell_height = DefaultCellHeight(newug);
        redisplay = TRUE;
    }

    if(curug->ugrid.foreground_pixel != newug->ugrid.foreground_pixel) {
        XtReleaseGC(new, curug->ugrid.text_gc);
        XtReleaseGC(new, curug->ugrid.selected_gc);
        newug->ugrid.text_gc = get_gc(newug, 
                                      newug->ugrid.foreground_pixel,
                                      newug->core.background_pixel);
        newug->ugrid.selected_gc = get_gc(newug, 
                                          newug->core.background_pixel,
                                          newug->ugrid.foreground_pixel);
        redisplay = TRUE;
    }

    if(curug->ugrid.start_codepoint != newug->ugrid.start_codepoint) {
        redisplay = TRUE;
    }

    return redisplay;
}

static void
Destroy(Widget w)
{
    UGridWidget ugw=(UGridWidget)w;
    XtReleaseGC(w, ugw->ugrid.text_gc);
    XtReleaseGC(w, ugw->ugrid.selected_gc);
    XtReleaseGC(w, ugw->ugrid.grid_gc);
}

static XtGeometryResult
QueryGeometry(Widget w, 
              XtWidgetGeometry *intended, XtWidgetGeometry *preferred)
{
    UGridWidget ugw=(UGridWidget)w;

    preferred->request_mode = CWWidth | CWHeight;
    preferred->width = DefaultCellWidth(ugw) * ugw->ugrid.cell_cols +
        ugw->ugrid.grid_width * (ugw->ugrid.cell_cols+1);
    preferred->height = DefaultCellHeight(ugw) * ugw->ugrid.cell_rows +
        ugw->ugrid.grid_width * (ugw->ugrid.cell_rows+1);

    if((intended->request_mode & (CWWidth | CWHeight)) == (CWWidth | CWHeight)
       && 
       intended->width == preferred->width &&
       intended->height == preferred->height)
        return XtGeometryYes;
    else if(preferred->width == w->core.width &&
            preferred->height == w->core.height)
        return XtGeometryNo;
    else
        return XtGeometryAlmost;
}

static void PageForward(Widget w, XEvent *event, 
                        String *params, Cardinal *nparams)
{
    UGridWidget ugw = (UGridWidget)w;
    UGridPart *p = &ugw->ugrid;
    int page = p->cell_cols*p->cell_rows;
    if(p->start_codepoint + page <= 0xFFFF)
        p->start_codepoint += page;
    else {
        XBell(XtDisplay(w), 0);
        return;
    }
    XClearWindow(XtDisplay(ugw), XtWindow(ugw));
    redisplay_rectangle(ugw, NULL);
    Notify(w, event, NULL, 0);
}
 
static void PageBackward(Widget w, XEvent *event, 
                         String *params, Cardinal *nparams)
{
    UGridWidget ugw = (UGridWidget)w;
    UGridPart *p = &ugw->ugrid;
    int page = p->cell_cols*p->cell_rows;

    if(p->start_codepoint - page > -page)
        p->start_codepoint -= page;
    else {
        XBell(XtDisplay(w), 0);
        return;
    }
    XClearWindow(XtDisplay(ugw), XtWindow(ugw));
    redisplay_rectangle(ugw, NULL);
    Notify(w, event, NULL, 0);
}

static long
coordinates_to_codepoint(Widget w, int x, int y)
{
    UGridWidget ugw = (UGridWidget)w;
    UGridPart *p = &ugw->ugrid;
    int hgw, cw, ch, i, j;
    long codepoint;

    hgw = p->grid_width/2;
    cw = p->cell_width+p->grid_width;
    ch = p->cell_height+p->grid_width;
    i = (x - hgw) / cw;
    j = (y - hgw) / ch;

    if(i < 0 || j < 0 || i >= p->cell_cols || j >= p->cell_rows)
        return 0;

    codepoint = p->start_codepoint + i + j*p->cell_cols;
    if(codepoint > 0xFFFFl)
        return 0;

    return codepoint;
}

static void
Notify(Widget w, XEvent *event, String *params, Cardinal *nparams)
{
    UGridWidget ugw = (UGridWidget)w;
    UGridPart *p = &ugw->ugrid;
    UGridCallbackDataRec rec;

    rec.start_codepoint = p->start_codepoint;
    rec.selected_codepoint = p->selected_codepoint;
  
    XtCallCallbacks(w, XtNcallback, (XtPointer)&rec);
}


static Boolean
codepoint_to_rectangle(UGridWidget ugw, XRectangle *rect,
                       long codepoint)
{
    UGridPart *p = &ugw->ugrid;
    int hgw, cw, ch, i, j, n;

    if(!CODEPOINT(codepoint)) {
        return False;
    }
    
    n = codepoint - p->start_codepoint;
    if(n < 0 || n > p->cell_rows * p->cell_cols)
        return False;
    j = n / p->cell_cols;
    i = n % p->cell_cols;
  
    hgw = p->grid_width/2;
    cw = p->cell_width + p->grid_width;
    ch = p->cell_height + p->grid_width;
    rect->x = hgw + i*cw;
    rect->y = hgw + j*ch;
    rect->width = cw;
    rect->height = ch;
    return True;
}

static void 
claim_selection(UGridWidget ugw,
                unsigned long codepoint, Time time)
{
    UGridPart *p = &ugw->ugrid;
    XRectangle rect;
    int page = p->cell_cols * p->cell_rows;
    Boolean redisplay = False;

    if(CODEPOINT(codepoint) && (codepoint < 0 || codepoint > 0xFFFF)) {
        claim_selection(ugw, NO_CODEPOINT, time);
        return;
    }

    if(CODEPOINT(codepoint) && 
       (codepoint / page) * page != p->start_codepoint) {
        p->start_codepoint = (codepoint / page) * page;
        redisplay = True;
    }

    if(!redisplay && 
       codepoint_to_rectangle(ugw, &rect, p->selected_codepoint)) {
        XClearArea(XtDisplay(ugw), XtWindow(ugw),
                   rect.x, rect.y, rect.width, rect.height, FALSE);
        p->selected_codepoint = NO_CODEPOINT;
        redisplay_rectangle(ugw, &rect);
    }

    p->selected_codepoint = codepoint;
    if(redisplay) {
        XClearWindow(XtDisplay(ugw), XtWindow(ugw));
        redisplay_rectangle(ugw, NULL);
    } else if(codepoint_to_rectangle(ugw, &rect, p->selected_codepoint)) {
        XClearArea(XtDisplay(ugw), XtWindow(ugw),
                   rect.x, rect.y, rect.width, rect.height, FALSE);
        redisplay_rectangle(ugw, &rect);
    }

    if(CODEPOINT(codepoint))
        XtOwnSelection((Widget)ugw, XA_PRIMARY, time, ConvertSelection,
                       LoseSelection, (XtSelectionDoneProc)NULL);
    Notify((Widget)ugw, NULL, NULL, 0);
}

static char*
codepoint_to_utf8(Dimension cp, unsigned long *length)
{
    static char c[4];
    
    if(cp > 0xFFFF)
        cp=0xFFFD;

    if(cp < 0x80) {
        c[0] = cp;
        *length = 1;
    } else if(cp < 0x800l) {
        c[0] = (0xC0 | cp>>6);
        c[1] = (0x80 | (cp&0x3F));
        *length = 2;
    } else if(cp < 0x10000l) {
        c[0]= (0xE0 | cp>>12);
        c[1]= (0x80 | (cp>>6 & 0x3F));
        c[2]= (0x80 | (cp & 0x3F));
        *length=3;
    } else
        *length=0;              /* this cannot happen */

    c[*length]='\0';
    return c;
}

static Boolean
utf8_to_codepoint(unsigned char *c, unsigned long length, long *cp_return)
{
    if(length < 1)
        return False;
    if((c[0]&0x80) == 0) {
        *cp_return = (c[0]&0x7F);
        return True;
    }

    if(length < 2)
        return False;
    if((c[0]&0xA0) == 0x80) {
        *cp_return = (c[0]&0x1F) << 6 | (c[1]&0x3F);
        return True;
    }
    
    if(length < 3)
        return False;
    if((c[0]&0x70) == 0x60) {
        *cp_return = (c[0] & 0x0F) << 12 |
                     (c[1] & 0x3F) << 6 |
                     (c[2] & 0x3F);
        return True;
    }

    return False;
}

static void 
SelectCodepoint(Widget w, XEvent *event, 
                String *params, Cardinal *nparams)
{
    int x, y;
    Time time;
    unsigned long codepoint;
    
    switch (event->type) {
    case KeyPress:
    case KeyRelease:
        x = event->xkey.x;
        y = event->xkey.y;
        time = event->xkey.time;
        break;
    case ButtonPress:
    case ButtonRelease:
        x = event->xbutton.x;
        y = event->xbutton.y;
        time = event->xbutton.time;
        break;
    default:
        return;
    }
  
    codepoint = coordinates_to_codepoint(w, x, y);
    if(CODEPOINT(codepoint))
        claim_selection((UGridWidget)w, codepoint, time);
                  
}

static void 
ClearSelection(Widget w, XEvent *event, 
               String *params, Cardinal *nparams)
{
    claim_selection((UGridWidget)w, NO_CODEPOINT, 0);
}

#ifndef X_HAVE_UTF8_STRING
static Atom XA_UTF8_STRING(Display *dpy)
{
    static AtomPtr p = NULL;
  
    if(p == NULL)
        p = XmuMakeAtom("UTF8_STRING");
    return XmuInternAtom(dpy, p);
}
#endif

static Boolean
ConvertSelection(Widget w, Atom *selection, Atom *target, Atom *type,
                 XtPointer *value, unsigned long *length, int *format)
{
    UGridWidget ugw=(UGridWidget)w;
    UGridPart *p=&ugw->ugrid;
    Display *dpy=XtDisplay(ugw);
    Atom myTarget;

    if(!CODEPOINT(p->selected_codepoint))
        return False;

    myTarget=*target;

    if(*selection != XA_PRIMARY)
        return False;

    if(myTarget == XA_TARGETS(dpy)) {
        Atom *targetP;
        *value = XtMalloc((unsigned)sizeof(Atom)*5);
        *length = 0;
        targetP = *(Atom**)value;
        *targetP++ = XA_TARGETS(dpy); (*length)++;
        *targetP++ = XA_STRING; (*length)++;
        *targetP++ = XA_TEXT(dpy); (*length)++;
        *targetP++ = XA_UTF8_STRING(dpy); (*length)++;
#ifdef X_HAVE_UTF8_STRING
        *targetP++ = XA_COMPOUND_TEXT(dpy); (*length)++;
#endif
        *type = XA_ATOM;
        *format = 32;
        return True;
    }

    if(myTarget == XA_TEXT(dpy)) {
        myTarget = XA_STRING;
#ifdef X_HAVE_UTF8_STRING
        if(p->selected_codepoint >= 0x100)
            myTarget = XA_COMPOUND_TEXT(dpy);
#endif
    }

    if(myTarget == XA_UTF8_STRING(dpy)) {
        char *c;
        c = codepoint_to_utf8(p->selected_codepoint, length);
        if(*length == 0)
            return False;
        *value = XtMalloc(*length);
        if(*value == NULL)
            return False;
        memcpy(*value, c, *length);
        *type = XA_UTF8_STRING(dpy);
        *format = 8;
        return True;
  }

#ifdef X_HAVE_UTF8_STRING
    if(myTarget == XA_COMPOUND_TEXT(dpy)) {
        int rc;
        XTextProperty prop;
        char *tl[2];
        unsigned long l;

        tl[0] = codepoint_to_utf8(p->selected_codepoint, &l);
        tl[1] = NULL;

        rc = Xutf8TextListToTextProperty(dpy, tl, 1, 
                                         XCompoundTextStyle, &prop);

        if(rc < 0)
            return False;

        if(prop.format != 8 || prop.nitems <= 0) {
            XFree(prop.value);
            return False;
        }
        *value = XtMalloc(prop.nitems);
        if(*value == NULL) {
            XFree(prop.value);
            return False;
        }

        memcpy(*value, prop.value, prop.nitems);
        XFree(prop.value);
        *length = prop.nitems;
        *type = prop.encoding;
        *format = prop.format;
        return True;
    }
#endif  

    if(myTarget == XA_STRING) {
        char c;
        if(p->selected_codepoint < 0x100)
            c = p->selected_codepoint;
        else
            c = '#';
        *value = XtMalloc(1);
        if(*value == NULL)
            return False;
        ((char*)*value)[0] = c;
        *length=1;
        *type = XA_STRING;
        *format = 8;
        return True;
    }

    return False;
}

static void
LoseSelection(Widget w, Atom *selection)
{
    UGridWidget ugw = (UGridWidget)w;

    if(*selection == XA_PRIMARY)
        claim_selection(ugw, NO_CODEPOINT, 0);
}

typedef struct _SelectionClientDataRec {
    Time time;
    Atom *targets;
} SelectionClientDataRec, *SelectionClientDataPtr;

static Atom*
SelectionTargets(Widget w)
{
#ifdef X_HAVE_UTF8_STRING
#define NUMTARGETS 3
#else
#define NUMTARGETS 2
#endif
    static Atom *targets = NULL;
    Display *dpy = XtDisplay(w);
    int n;

    if(!targets) {
        targets = (Atom*)XtMalloc((NUMTARGETS + 1) * sizeof(Atom));
        if(!targets)
            return NULL;
        n = 0;
        targets[n] = XA_UTF8_STRING(dpy); n++;
#ifdef X_HAVE_UTF8_STRING
        targets[n] = XA_COMPOUND_TEXT(dpy); n++;
#endif
        targets[n] = XA_STRING; n++;
        targets[n] = None;
    }

    return targets;
}

static void 
InsertSelection(Widget w, XEvent *event, 
                String *params, Cardinal *nparams)
{
    Time time;
    SelectionClientDataPtr cd;
    Atom target, *targets;

    switch (event->type) {
    case KeyPress:
    case KeyRelease:
        time = event->xkey.time;
        break;
    case ButtonPress:
    case ButtonRelease:
        time = event->xbutton.time;
        break;
    default:
        return;
  }

    cd = (SelectionClientDataPtr)XtMalloc(sizeof(SelectionClientDataRec));
    targets = SelectionTargets(w);
    target = targets[0];
    cd->time = time;
    cd->targets = targets + 1;
              
    XtGetSelectionValue(w, XA_PRIMARY, target,
                        SelectionReceived, (XtPointer)cd, time);
}

static void
SelectionReceived(Widget w, XtPointer client_data, 
                  Atom *selection, Atom *type,
                  XtPointer value, unsigned long *length, int *format)
{
    long codepoint;
    SelectionClientDataPtr cd = (SelectionClientDataPtr)client_data;
    Boolean ret;

    if(*type == XA_UTF8_STRING(XtDisplay(w))) {
        if(*format != 8 || *length < 1)
            goto fail;

        ret = utf8_to_codepoint((unsigned char*)value, *length, &codepoint);
        if(!ret)
            goto fail;

        claim_selection((UGridWidget)w, codepoint, cd->time);
        XtFree((XtPointer)cd);
        return;
    }

#ifdef X_HAVE_UTF8_STRING
    if(*type == XA_COMPOUND_TEXT(XtDisplay(w))) {
        XTextProperty prop;
        char **list;
        int rc;
        int count;

        if(*format != 8 || *length < 1)
            goto fail;

        prop.value = (unsigned char*)value;
        prop.encoding = XA_COMPOUND_TEXT(XtDisplay(w));
        prop.format = 8;
        prop.nitems = *length;
        
        rc = Xutf8TextPropertyToTextList(XtDisplay(w),
                                         &prop,
                                         &list,
                                         &count);
        if(rc < 0)
            goto fail;
        if(count <= 0) {
            XFree(list);
            goto fail;
        }

        ret = utf8_to_codepoint((unsigned char*)list[0], strlen(list[0]), 
                              &codepoint);
        XFree(list);
        if(!ret)
            goto fail;

        claim_selection((UGridWidget)w, codepoint, cd->time);
        XtFree((XtPointer)cd);
        return;
  }
#endif

    if(*type == XA_STRING) {
        if(*format !=8 || *length < 1)
            goto fail;
        codepoint = (((unsigned char*)value)[0]&0xFF);
        claim_selection((UGridWidget)w, codepoint, cd->time);
        XtFree((XtPointer)cd);
        return;
    }
  
fail:
    if(cd->targets[0] == None) {
        XtFree((XtPointer)cd);
        XBell(XtDisplay(w), 0);
        return;
    } else {
        Atom target;
        target = cd->targets[0];
        cd->targets = cd->targets + 1;

        XtGetSelectionValue(w, XA_PRIMARY, target,
                            SelectionReceived, (XtPointer)cd, cd->time);
        return;
    }
}
